/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.storage;

import java.util.Calendar;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.ItemExistsException;
import javax.jcr.RepositoryException;

import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;

/**
 * Created by The eXo Platform SARL .<br>
 * The rules are: - data retrieving by accessors (getItem, getNodeByUUID) do not
 * have to (but can) be fully initialized - data sending to add, update mutators
 * have be fully initialized - data sendig to delete do not have to be fully
 * initialized
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: WorkspaceDataContainer.java 3510 2005-12-29 17:32:57Z geaz $
 */
//An InvalidItemStateException may be thrown on a write method of an Item if
//the change being made would (upon save) conflict with a change made, and
//persisted, through another Session. If detection of the conflict is only
// possible
//at save-time, then save will throw an InvalidItemStateException.
public interface WorkspaceDataContainer {

  /**
   * @return name of this container (see container configuration)
   */
  String getName();

  /**
   * @return current time as for this container env
   */
  Calendar getCurrentTime();

  /**
   * @param loc -
   *          item's location
   * @return - stored PropertyData wich has exact the same path inside location;
   *         or stored NodeData which has exact the same path/index if index is
   *         initialized inside location or index=1 otherwise; null if such a
   *         location not found
   */
  ItemData getItem(InternalQPath qPath);

  /**
   * @param uuid -
   *          unique identifier
   * @return corresponding stored ItemData or null. Basically used for
   *         Session.getNodeByUUID but not necessarily refers to jcr:uuid
   *         property (In fact, this uuid should not necessary be equal of
   *         referenceable node's UUID if any) thereby can return NodeData for
   *         not referenceable node or PropertyData.
   */
  ItemData getItemByUUID(String uuid);

  //NodeData getNodeByUUID(String uuid);

  /**
   * Adds NodeData. All descendants should be added as well As well responsible
   * for maintaining binding to its parent's NodeData - i.e. next get method
   * that returns its parent should have this item as a child.
   * 
   * @param item
   * @return - stored NodeData
   * @throws ItemExistsException -
   *           if NodeData already exists on this location
   * @throws RepositoryException
   */
  NodeData add(NodeData data) throws ItemExistsException, RepositoryException,
      InvalidItemStateException;

  /**
   * Adds single property data
   * 
   * @param data
   * @return
   * @throws ItemExistsException
   * @throws RepositoryException
   * @throws InvalidItemStateException
   */
  PropertyData add(PropertyData data) throws ItemExistsException,
      RepositoryException, InvalidItemStateException;

  /**
   * Updates stored NodeData. Updating node operations includes reordering
   * (Node.orderBefore) As well can be responsible (depending on storage
   * mechanism) for maintaining binding to its parent's NodeData - i.e. next get
   * method that returns its parent should have this item as a child.
   * 
   * @param item
   * @return stored PropertyData
   * @throws RepositoryException
   */
  NodeData update(NodeData data) throws RepositoryException,
      InvalidItemStateException;

  /**
   * Updates stored or Adds new PropertyData. As well responsible for
   * maintaining binding to its parent's NodeData - i.e. next get method that
   * returns its parent should have this item as a child.
   * 
   * @param item
   * @return stored PropertyData
   * @throws RepositoryException
   */
  PropertyData update(PropertyData data) throws RepositoryException,
      InvalidItemStateException;

  /**
   * Deletes node or property data
   * 
   * @param data
   * @throws RepositoryException
   */
  void delete(ItemData data) throws RepositoryException,
      InvalidItemStateException;

  /**
   * @param UUID
   * @return list of
   * @throws RepositoryException
   */
  List getReferences(String UUID) throws RepositoryException;

  /**
   * @return info about this container
   */
  String getInfo();
  
  void beginTransaction();

  void commit();

  void rollback();
  
  void close();

}