/**
 **************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jcr.Credentials;
import javax.jcr.LoginException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.Node;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.SimpleCredentials;
import javax.jcr.nodetype.NodeTypeManager;
import javax.security.auth.login.LoginContext;

import org.apache.commons.logging.Log;
import org.exoplatform.container.SessionContainer;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.access.SystemIdentity;
import org.exoplatform.services.jcr.config.RepositoryConfigurationException;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.core.ManageableRepository;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.jcr.datamodel.IllegalPathException;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.RepositoryContainer;
import org.exoplatform.services.jcr.impl.WorkspaceContainer;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.storage.TransientNodeData;
import org.exoplatform.services.jcr.impl.storage.TransientPropertyData;
import org.exoplatform.services.jcr.impl.storage.TransientValueData;
import org.exoplatform.services.jcr.storage.WorkspaceDataContainer;
import org.exoplatform.services.log.LogService;
import org.exoplatform.services.security.impl.mock.MockCallbackHandler;
import org.picocontainer.ComponentAdapter;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: RepositoryImpl.java 3551 2006-01-03 09:07:25Z lautarul $
 */
public class RepositoryImpl implements ManageableRepository {

  private static HashMap descriptors = new HashMap();
  static {
    descriptors.put(SPEC_VERSION_DESC, "1.0");
    descriptors.put(SPEC_NAME_DESC, "Content Repository Java Technology API");
    descriptors.put(REP_VENDOR_DESC, "eXo Platform SARL");
    descriptors.put(REP_VENDOR_URL_DESC, "http://www.exoplatform.com");
    descriptors.put(REP_NAME_DESC, "eXo Java Content Repository");
    descriptors.put(REP_VERSION_DESC, "1.0-RC-2");
    descriptors.put(LEVEL_1_SUPPORTED, "true");
    descriptors.put(LEVEL_2_SUPPORTED, "true");
    descriptors.put(OPTION_TRANSACTIONS_SUPPORTED, "false");
    descriptors.put(OPTION_VERSIONING_SUPPORTED, "true");
    descriptors.put(OPTION_OBSERVATION_SUPPORTED, "true");
    descriptors.put(OPTION_LOCKING_SUPPORTED, "true");
    descriptors.put(OPTION_QUERY_SQL_SUPPORTED, "true");
    descriptors.put(QUERY_XPATH_POS_INDEX, "true");
    descriptors.put(QUERY_XPATH_DOC_ORDER, "true");
  }

  private RepositoryContainer repositoryContainer;

  private String systemWorkspaceName;

  protected Log log;

  private LogService logService;

  private String name;
  
  private String accessControlType;

  private RepositoryEntry config;

  public RepositoryImpl(RepositoryContainer container)
      throws RepositoryException, RepositoryConfigurationException {

    this.logService = (LogService) container
        .getComponentInstanceOfType(LogService.class);

    this.log = logService.getLog("org.exoplatform.services.jcr");
    
    config = (RepositoryEntry) container
        .getComponentInstanceOfType(RepositoryEntry.class);

    this.name = config.getName();
    this.systemWorkspaceName = config.getSystemWorkspaceName();
    this.repositoryContainer = container;
    this.accessControlType = config.getAccessControl();
  }

  /**
   * @see javax.jcr.Repository#getDescriptorKeys
   */
  public String[] getDescriptorKeys() {
    String[] keys = new String[descriptors.size()];
    Iterator decriptorsList = descriptors.keySet().iterator();
    int i = 0;
    while (decriptorsList.hasNext())
      keys[i++] = (String) decriptorsList.next();
    return keys;
  }

  /**
   * @see javax.jcr.Repository#getDescriptor
   */
  public String getDescriptor(String key) {
    return (String) descriptors.get(key);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(Credentials credentials) throws LoginException,
      NoSuchWorkspaceException, RepositoryException {
    return login(credentials, null);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(String workspaceName) throws LoginException,
      NoSuchWorkspaceException, RepositoryException {
    return login(null, workspaceName);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login() throws LoginException, NoSuchWorkspaceException,
      RepositoryException {
    return login(null, null);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(Credentials credentials, String workspaceName)
      throws LoginException, NoSuchWorkspaceException, RepositoryException {

    if (workspaceName == null) {
      if (systemWorkspaceName == null)
        throw new NoSuchWorkspaceException(
            "Both workspace and default-workspace name are null! ");
      else
        workspaceName = systemWorkspaceName;
    }

    if (!isWorkspaceInitialized(workspaceName))
      throw new NoSuchWorkspaceException(
          "Workspace '"
              + workspaceName
              + "' not found. "
              + "Probably is not initialized. If so either Initialize it manually or turn on the RepositoryInitializer");

    CredentialsImpl thisCredentials;
    // 6.1.2 The implementor may either provide its own custom
    // implementation or use the provided
    // SimpleCredentials class. This class provides a minimal standard
    // method for authenticating
    // against a repository (i.e., using a user ID and password).
    if (credentials != null) {
      if (credentials instanceof CredentialsImpl) {
        thisCredentials = (CredentialsImpl) credentials;
        authenticate(thisCredentials); // throws LoginException
      } else if (credentials instanceof SimpleCredentials) {
        String name = ((SimpleCredentials) credentials).getUserID();
        char[] pswd = ((SimpleCredentials) credentials).getPassword();
        thisCredentials = new CredentialsImpl(name, pswd);
        authenticate(thisCredentials); // throws LoginException
      } else
        throw new LoginException(
            "Credentials for the authentication should be CredentialsImpl or SimpleCredentials type");
      log.debug("Repository.login() authenticated "
          + thisCredentials.getUserID());

    } else {
      SessionContainer sessionContainer = SessionContainer.getInstance();
      if (sessionContainer != null) {
        String userId = SessionContainer.getInstance().getRemoteUser();
        if (userId == null)
          userId = SystemIdentity.ANONIM;
        thisCredentials = new CredentialsImpl(userId, new char[0]);

        log.debug("Repository.login() gets user  " + userId
            + " from SessionContainer");
      } else {
        throw new LoginException(
            "Current User should be authenticated externally but was not.");
      }

    }
    WorkspaceContainer workspaceContainer = (WorkspaceContainer) repositoryContainer
        .getComponentInstance(workspaceName);
    
    //SessionFactory factory = (SessionFactory)workspaceContainer.
    //getComponentInstanceOfType(SessionFactory.class);
    //return factory.createSession(thisCredentials, workspaceName);

    return new SessionImpl(workspaceName, thisCredentials, workspaceContainer);

  }
  

  /// -------- ManageableRepository impl -----------------------------
  
  public Session getSystemSession(String workspaceName)
      throws RepositoryException {
    WorkspaceContainer wsContainer = (WorkspaceContainer) repositoryContainer
        .getComponentInstance(workspaceName);
    if (wsContainer == null)
      throw new RepositoryException("No workspace container configured for "
          + workspaceName + " @ " + this.name);
    return (Session) wsContainer.getComponentInstanceOfType(Session.class);
  }

  public RepositoryEntry getConfiguration() {
    return config;
  }
  
  public String[] getWorkspaceNames() {
    List adapters = repositoryContainer
        .getComponentAdaptersOfType(WorkspaceContainer.class);
    String[] workspaceNames = new String[adapters.size()];
    for (int i = 0; i < adapters.size(); i++) {
      ComponentAdapter adapter = (ComponentAdapter) adapters.get(i);
      workspaceNames[i] = new String((String) adapter.getComponentKey());
    }
    return workspaceNames;
  }

  public ExtendedNodeTypeManager getNodeTypeManager() {
    
    return (ExtendedNodeTypeManager) repositoryContainer
        .getComponentInstanceOfType(NodeTypeManager.class);
  }
  
  public NamespaceRegistry getNamespaceRegistry() {
    
    return (NamespaceRegistry) repositoryContainer
        .getComponentInstanceOfType(NamespaceRegistry.class);
  }
  

  public void initWorkspace(String workspaceName, String rootNodeType)
      throws RepositoryException {

    if (isWorkspaceInitialized(workspaceName)) {
      log.warn("Workspace '" + workspaceName
          + "' is presumably initialized. Initialization canceled");
      return;
    }

    if (!isWorkspaceInitialized(systemWorkspaceName)
        && !workspaceName.equals(systemWorkspaceName))
      log.warn("System workspace is not initialized for repository " + name);

    // Init root (/) node
    Node root = initRootNode(workspaceName, rootNodeType);

    log.info("Root node for " + workspaceName + " initialized. NodeType: "
        + rootNodeType + " system workspace: " + systemWorkspaceName);

    // Init system workspace
    if (workspaceName.equals(systemWorkspaceName)) {
      // initialize /jcr:system
      Node sys = initJcrSystemNode(root);
    }

    log.info("Workspace " + workspaceName + " @ " + this.name
        + " is initialized");

  }

  public boolean isWorkspaceInitialized(String workspaceName) {
    try {
      SessionImpl session = (SessionImpl) getSystemSession(workspaceName);
      session.getItem(JCRPath.ROOT_PATH);
      // jcr:system should be initialized as well
      if (workspaceName.equals(systemWorkspaceName))
        session.getItem(JCRPath.SYSTEM_PATH);
      return true;
    } catch (Exception e) {
      //e.printStackTrace();
      return false;
    }
  }

  ////////////////////////////////////////////////////////

  public RepositoryContainer getContainer() {
    return repositoryContainer;
  }

  public String getSystemWorkspaceName() {
    return systemWorkspaceName;
  }

  public String getName() {
    return name;
  }
  

  private void authenticate(CredentialsImpl credentials)
      throws LoginException {
    
    // SYSTEM
    if(credentials.getUserID().equals(SystemIdentity.SYSTEM))
      return;

    MockCallbackHandler handler = new MockCallbackHandler(credentials
        .getUserID(), credentials.getPassword());
    try {
      LoginContext loginContext = new LoginContext("eXo", handler);
      loginContext.login();
    } catch (javax.security.auth.login.LoginException e) {
      throw new LoginException("Login failed for "+credentials.getUserID()+ " " + e);
    }
    //System.out.println("Logged "+credentials.getUserID());

  }

  public SessionImpl getSystemSession() throws RepositoryException {
    return (SessionImpl) getSystemSession(systemWorkspaceName);
  }

  private WorkspaceContainer getWorkspaceContainer(String workspaceName)
      throws RepositoryException {

    WorkspaceContainer wsc = (WorkspaceContainer) repositoryContainer
        .getComponentInstance(workspaceName);
    if (wsc == null)
      throw new RepositoryException("No container found for workspace '"
          + workspaceName + "' at " + this.name);
    return wsc;
  }

  private LocationFactory getLocationFactory() {
    return (LocationFactory) repositoryContainer
        .getComponentInstanceOfType(LocationFactory.class);

  }

  private Node initRootNode(String workspaceName, String rootNodeType)
      throws RepositoryException {

    JCRPath rootLocation = getLocationFactory().createRootLocation();
    JCRName ntName = getLocationFactory().parseJCRName(rootNodeType);

    TransientNodeData nodeData = new TransientNodeData(rootLocation
        .getInternalPath(), UUIDGenerator.ROOT_UUID, -1, ntName
        .getInternalName(), new InternalQName[0], 0, null);

    UUIDGenerator uuidGenerator = (UUIDGenerator) repositoryContainer
        .getComponentInstanceOfType(UUIDGenerator.class);
    try {
      TransientPropertyData primaryType = new TransientPropertyData(
          InternalQPath.makeChildPath(nodeData.getQPath(),
              Constants.PRIMARY_TYPE_URI), uuidGenerator
              .generateStringUUID(Constants.PRIMARY_TYPE_URI), -1,
              PropertyType.NAME, nodeData.getUUID()); 
      primaryType.setValue(new TransientValueData(getLocationFactory()
          .parseJCRName(rootNodeType).getInternalName()));

      nodeData.addProperty(primaryType);
    } catch (IllegalPathException e) {

    }
    WorkspaceDataContainer wsContainer = (WorkspaceDataContainer) getWorkspaceContainer(
        workspaceName).getComponentInstanceOfType(WorkspaceDataContainer.class);

   
    wsContainer.beginTransaction();
    wsContainer.add(nodeData);
    wsContainer.commit();

    SessionImpl session = (SessionImpl) getSystemSession(workspaceName);
    return (Node) session.getItem(JCRPath.ROOT_PATH);
  }

  private Node initJcrSystemNode(Node root) throws RepositoryException {
    NodeImpl sys = (NodeImpl) root.addNode("jcr:system", "nt:unstructured");
    
    if (!accessControlType.equals(AccessControlPolicy.DISABLE)) {
      HashMap perm = new HashMap();
      perm.put("*:/admin", new String[] { PermissionType.READ });
      sys.setChildrenPermissions(perm);
    }
    
    WorkspaceImpl ws = (WorkspaceImpl)root.getSession().getWorkspace();


    ((NamespaceRegistryImpl) ws.getNamespaceRegistry()).initStorage(sys, accessControlType);
    ((NodeTypeManagerImpl) getNodeTypeManager()).initStorage(sys);
    ws.getVersionStorageManager().initStorage(sys, accessControlType);

    //	    log.debug("System DataManager: \n"+((SessionImpl)root.getSession()).getTransientNodesManager().dump());

    root.save();

    return sys;
  }

}