/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.storage.TransientValueData;


/**
 * A <code>BinaryValue</code> provides an implementation
 * of the <code>Value</code> interface representing a binary value.
 *
 * @author Stefan Guggisberg
 */
public class BinaryValue extends BaseValue {

    public static final int TYPE = PropertyType.BINARY;

    // those fields are mutually exclusive, i.e. only one can be non-null
    //private byte[] streamData = null;
    //private String text = null;

    /**
     * Constructs a <code>BinaryValue</code> object based on a string.
     *
     * @param text the string this <code>BinaryValue</code> should represent
     */
    public BinaryValue(String text) {
        super(TYPE, new TransientValueData(text));
    }

    public BinaryValue(byte[] bytes) {
        super(TYPE, new TransientValueData(bytes));
    }

    /**
     * Constructs a <code>BinaryValue</code> object based on a stream.
     *
     * @param stream the stream this <code>BinaryValue</code> should represent
     */
    public BinaryValue(InputStream stream) {
        super(TYPE, null);
        
        byte[] buf;
		try {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            byte[] buffer = new byte[8192];
            int read;
            while((read = stream.read(buffer)) > 0) {
                out.write(buffer, 0, read);
            }
            buf = out.toByteArray();
			this.data = new TransientValueData(buf);
	    	this.stream = new ByteArrayInputStream(buf);
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException("Initializing BinaryValue error "+e);
			//throw new RepositoryException(e);
		}
    }

    /**
     * Constructs a <code>BinaryValue</code> object based on a stream.
     *
     * @param data the stream this <code>BinaryValue</code> should represent
     */
    public BinaryValue(ValueData data) {
        super(TYPE, data);
    }

    /**
     * Indicates whether some other object is "equal to" this one.
     * <p>
     * The result is <code>true</code> if and only if the argument is not
     * <code>null</code> and is a <code>BinaryValue</code> object that
     * represents the same value as this object.
     *
     * @param   obj   the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     */
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj instanceof BinaryValue) {
            BinaryValue other = (BinaryValue) obj;
            if (stream != null) {
                return stream.equals(other.stream);
            } else if (data != null) {
                return data.equals(other.getData());
            } 
        }
        return false;
    }

    /**
     * Gets the string representaion of this binary value.
     */
    public String getInternalString() throws ValueFormatException, RepositoryException {
        // build text value if necessary
    	String text = null;
        if (data != null) {
            try {
                text = new String(data.getAsByteArray(), DEFAULT_ENCODING);
            } catch (UnsupportedEncodingException e) {
                throw new RepositoryException(DEFAULT_ENCODING + " not supported on this platform", e);
            }
        } else if (stream != null) {
            try {
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                byte[] buffer = new byte[8192];
                int read;
                while ((read = stream.read(buffer)) > 0) {
                    out.write(buffer, 0, read);
                }
                byte[] data = out.toByteArray();
                text = new String(data, DEFAULT_ENCODING);
            } catch (UnsupportedEncodingException e) {
                throw new RepositoryException(DEFAULT_ENCODING + " not supported on this platform", e);
            } catch (IOException e) {
                throw new ValueFormatException("conversion from stream to string failed", e);
            } finally {
                try {
                    if (stream != null) {
                        stream.close();
                    }
                } catch (IOException e) {
                    // ignore
                }
            }
            stream = null;
        }

        if (text != null) {
            return text;
        } else {
            throw new ValueFormatException("empty value");
        }
    }

    //----------------------------------------------------------------< Value >
    /**
     * @see Value#getStream
     */

    public InputStream getStream() throws ValueFormatException, IllegalStateException, RepositoryException {
        setStreamConsumed();

        // build stream value if necessary
        if (data != null) {
            stream = new ByteArrayInputStream(data.getAsByteArray());
        }

        return super.getStream();
    }
    
}
