/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.util.Calendar;
import java.util.Date;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.storage.TransientValueData;


/**
 * A <code>LongValue</code> provides an implementation
 * of the <code>Value</code> interface representing a long value.
 *
 * @author Stefan Guggisberg
 */
public class LongValue extends BaseValue {

    public static final int TYPE = PropertyType.LONG;

    private final Long lNumber;


    /**
     * Constructs a <code>LongValue</code> object representing a long.
     *
     * @param l the long this <code>LongValue</code> should represent
     */
    LongValue(long l) {
		super(TYPE, new TransientValueData(l));
		this.lNumber = new Long(l);
	}
    
    LongValue(ValueData data) {
		super(TYPE, data);
		this.lNumber = new Long(new String(data.getAsByteArray()));
	}

    /**
     * Indicates whether some other object is "equal to" this one.
     * <p/>
     * The result is <code>true</code> if and only if the argument is not
     * <code>null</code> and is a <code>LongValue</code> object that
     * represents the same value as this object.
     *
     * @param obj the reference object with which to compare.
     * @return <code>true</code> if this object is the same as the obj
     *         argument; <code>false</code> otherwise.
     */
    public boolean equals(Object obj) {
	if (this == obj) {
	    return true;
	}
	if (obj instanceof LongValue) {
	    LongValue other = (LongValue) obj;
	    if (lNumber == other.lNumber) {
		return true;
	    } else if (lNumber != null && other.lNumber != null) {
		return lNumber.equals(other.lNumber);
	    }
	}
	return false;
    }

    /**
     * @see BaseValue#getInternalString()
     */
    protected String getInternalString() throws ValueFormatException {
	if (lNumber != null) {
	    return lNumber.toString();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    //----------------------------------------------------------------< Value >
    /**
     * @see Value#getDate
     */
    public Calendar getDate() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (lNumber != null) {
	    // loosing timezone information...
	    Calendar cal = Calendar.getInstance();
	    cal.setTime(new Date(lNumber.longValue()));
	    return cal;
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    /**
     * @see Value#getLong
     */
    public long getLong() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (lNumber != null) {
	    return lNumber.longValue();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    /**
     * @see Value#getBoolean
     */
    public boolean getBoolean() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	throw new ValueFormatException("conversion to boolean failed: inconvertible types");
    }

    /**
     * @see Value#getDouble
     */
    public double getDouble() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (lNumber != null) {
	    return lNumber.doubleValue();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }
}
